/*
 * Copyright (c) 1999 G. Adam Stanislav
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	hutrans.c v.1, Unix version
 */

#include <stdio.h>
#include <stdlib.h>
#include <utf-8.h>

/*
 * Note: If you do not have <utf-8.h>, download libutf-8-1.0.tar.gz from
 * http://www.whizkidtech.net/i18n/
 *
 * The library enclosed in that package (libutf-8.so) is necessary for uhtrans
 * to work.
 */

#define	BUFFERSIZE	128

int inbuffer = 0;
int nextchar = 0;
static FILE *input   = stdin;
static FILE *output  = stdout;
static char buffer[BUFFERSIZE];

/*
 * getuchar() is actually a simple lexical analyzer.
 * It could be written in lex, but that would,
 * no doubt, produce code too bloated for our needs.
 */
static unsigned int getuchar(void) {
	unsigned int c;
	int hex;

	if (inbuffer) {
		c = buffer[nextchar++];
		if (nextchar == inbuffer)
			nextchar = inbuffer = 0;
		return c;
	}

	c = fgetc(input);
	if (c != '&')
		return c;

	c = fgetc(input);
	if (c != '#') {
		/*
		 * Resist the temptation of "optimizing" this
		 * by storing c in the buffer. This would cause
		 * subtle problems in two cases:
		 *
		 * (1) c == EOF - it would truncate EOF to (char)0xFF;
		 * (2) c == '&' - we could miss the next &#1234; sequence.
		 */
		ungetc(c, input);
		return '&';
	}

	buffer[inbuffer++] = '#';

	c = fgetc(input);

	if ((c == 'x') || (c == 'X'))
		hex = 1;
	else if (isdigit(c))
		hex = 0;
	else {
		/* See the note about temptation above */
		ungetc(c, input);
		return '&';
	}

	for (buffer[inbuffer++] = c; inbuffer < BUFFERSIZE; inbuffer++) {
		c = fgetc(input);
		if (isdigit(c) || (hex && isxdigit(c)))
			buffer[inbuffer] = c;
		else if ((c != ';') || (hex && (inbuffer == 2))) {
			/* See the note about temptation above */
			ungetc(c, input);
			return '&';
		}
		else {
			buffer[inbuffer] = '\0';
			inbuffer = 0;
			return strtoul(buffer + hex + 1, NULL, hex ? 16 : 10);
		}
	}

	/*
	 * If we get here, we have read &# followed
	 * by 127 digits. We'll just copy it all to
	 * the output...
	 */
	return '&';
}

int usage() {
	fprintf(stderr, "Usage: hutrans [ -i inputfile][ -o outputfile]\n");
	return 1;
}

int cantopen(char *file) {
	fprintf(stderr, "hutrans: Can't open %s.\n", file);
	if ((input != NULL) && (input != stdin))
		fclose(input);
	if ((output != NULL) && (output != stdout))
		fclose(output);
	return 2;
}

int main(int argc, char *argv[]) {
	int i;
	unsigned int c;
	char *inp  = NULL;
	char *outp = NULL;

	for (i = 1; i < argc; i++) {
		if (argv[i][0] != '-')
			return usage();
		switch (argv[i][1]) {
			case 'i':
				if (++i < argc)
					inp = argv[i];
				else return usage();
				break;
			case 'o':
				if (++i < argc)
					outp = argv[i];
				else return usage();
				break;
			default:
				return usage();
		}
	}

	if (inp != NULL) {
		input = fopen(inp, "r");
		if (input == NULL)
			return cantopen(inp);
	}

	if (outp != NULL) {
		output = fopen(outp, "w");
		if (output == NULL)
			return cantopen(outp);
	}

	for (c = getuchar(); c != (unsigned int)EOF; c = getuchar())
		utf8fputc(c, output);

	if (input != stdin)
		fclose(input);

	if (output != stdout)
		fclose(output);

	return 0;
}
