/*
 * Copyright (c) 1999 G. Adam Stanislav
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	utrans.c v.1, Unix version
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <utf-8.h>

/*
 * Note: If you do not have <utf-8.h>, download libutf-8-1.0.tar.gz from
 * http://www.whizkidtech.net/i18n/
 *
 * The library enclosed in that package (libutf-8.so) is necessary for utrans
 * to work.
 */

extern unsigned int utranstable[];
static int verbose = 0;
static int binary  = 1;
static unsigned int line = 1;
static unsigned int character = 1;
static FILE *input   = stdin;
static FILE *output  = stdout;
static FILE *page    = NULL;
static char *charmap = NULL;
static char buffer[9];

int usage() {
	fprintf(stderr, "Usage: utrans [-p pagefile][ -i inputfile][ -o outputfile][ -v][ -e][ -b|t]\n");
	return 1;
}

int cantopen(char *file) {
	fprintf(stderr, "utrans: Can't open %s.\n", file);
	if (charmap != NULL)
		free(charmap);
	if ((input != NULL) && (input != stdin))
		fclose(input);
	if ((output != NULL) && (output != stdout))
		fclose(output);
	/* No need to close page file, it is always NULL at this point */
	return 2;
}

int nem(void) {
	fprintf(stderr, "utrans: Not enough memory.\n");
	if ((input != NULL) && (input != stdin))
		fclose(input);
	if ((output != NULL) && (output != stdout))
		fclose(output);
	return 3;
}

void parse(void) {
	int c;
	int chars;
	unsigned int pagechar;
	unsigned int unichar;

	if (binary != 0) {
		if (verbose)
			fprintf(stderr, "utrans: Loading binary charmap.\n");
		fread(utranstable, sizeof(unsigned int), 256, page);
		return;
	}

	c = 0;

	while (c != EOF) {
		c = fgetc(page);
		if ((c == '=') || ((c == '/') && (fgetc(page) == 'x'))) {
			for (chars = 0; chars < 2; chars++) {
				c = fgetc(page);
				if (isxdigit(c))
					buffer[chars] = c;
				else {
					ungetc(c, page);
					break;
				}
			}
			buffer[chars] = '\0';
			pagechar = strtoul(buffer, NULL, 16);
			for (;;) {
				c = fgetc(page);
				if (c == EOF)
					return;
				else if (c == 'U') {
					c = fgetc(page);
					if (c != '+') {
						ungetc(c, page);
						if (!isxdigit(c)) continue;
					}

					for (chars = 0; chars < 8; chars++) {
						c = fgetc(page);
						if (isxdigit(c))
							buffer[chars] = c;
						else {
							ungetc(c, page);
							break;
						}
					}
					buffer[chars] = '\0';
					unichar = strtoul(buffer, NULL, 16);
					if (pagechar < 256) {
						utranstable[pagechar] = unichar;
						if (verbose)
							fprintf(stderr, "utrans: Converting %02x to %04x.\n", pagechar, unichar);
					}
					else if (verbose)
						fprintf(stderr, "utrans: Ignoring invalid character %02x.\n", pagechar);
					break;
				}
			}
		}
	}
}

int main(int argc, char *argv[]) {
	int i, j;
	int ignoreenvironment = 0;
	char *inp  = NULL;
	char *outp = NULL;
	char *pg   = NULL;
	char *cp;

	for (i = 1; i < argc; i++) {
		if (argv[i][0] != '-')
			return usage();
		switch (argv[i][1]) {
			case 'b':
				binary = 1;
				break;
			case 't':
				binary = 0;
				break;
			case 'p':
				if (++i < argc)
					pg = argv[i];
				else return usage();
				break;
			case 'i':
				if (++i < argc)
					inp = argv[i];
				else return usage();
				break;
			case 'o':
				if (++i < argc)
					outp = argv[i];
				else return usage();
				break;
			case 'e':
				ignoreenvironment = 1;
				break;
			case 'v':
				verbose = 1;
				break;
			default:
				return usage();
		}
	}

	cp = ignoreenvironment == 0 ? getenv("CHARMAPS") : NULL;

	if ((pg == NULL) && (ignoreenvironment == 0))
		pg = getenv("UTRANS");

	if (inp != NULL) {
		input = fopen(inp, "r");
		if (input == NULL)
			return cantopen(inp);
	}

	if (outp != NULL) {
		output = fopen(outp, "w");
		if (output == NULL)
			return cantopen(outp);
	}

	if (pg != NULL) {
		if (cp == NULL) {
			charmap = strdup(pg);
			if (charmap == NULL)
				return nem();
		}
		else {
			charmap = (char *)malloc(sizeof(pg) + sizeof(cp) + 2);
			if (charmap == NULL)
				return nem();
			sprintf(charmap, "%s/%s", cp, pg);
		}
		page = fopen(charmap, "r");
		if (page == NULL)
			return cantopen(charmap);
		else if (charmap != NULL) {
			free(charmap);
			charmap = NULL;
		}
	}

	if (page != NULL) {
		parse();
		fclose(page);
	}

	while ((i = fgetc(input)) != EOF) {
		if (i == '\n') {
			line++;
			character = 1;
		}
		j = utf8fputc(utranstable[i], output);
		if (j == EOF) {
			fprintf(stderr, "utrans (line %u, character %u): Error writing output (%02x, U+%04x, %c).\n", line, character, i, utranstable[i], i);
			break;
		}
		else if (j == 1)
			fprintf(stderr, "utrans (line %u, character %u): Error translating character %02x (U+%04x, %c).\n", line, character, i, utranstable[i], i);
		character++;
	}

	if ((input != NULL) && (input != stdin))
		fclose(input);
	if ((output != NULL) && (output != stdout))
		fclose(output);

	return 0;

}

